import redis

from django.utils.encoding import force_text

from mayan.apps.dependencies.exceptions import DependenciesException

from ..exceptions import LockError
from ..settings import setting_backend_arguments

from .base import LockingBackend
from .literals import REDIS_LOCK_VERSION_REQUIRED

SCAN_KEYS_COUNT = 5000
REDIS_LOCK_NAME_PREFIX = '_mayan_lock:'


class RedisLock(LockingBackend):
    @classmethod
    def _acquire_lock(cls, name, timeout):
        return RedisLock(name=name, timeout=timeout)

    @classmethod
    def get_redis_connection(cls):
        redis_url = setting_backend_arguments.value.get('redis_url', None)
        server = redis.from_url(url=redis_url)
        # Force to initialize the connection
        server.client()
        return server

    @classmethod
    def _purge_locks(cls):
        cache = cls.get_redis_connection()

        cursor = '0'
        while True:
            cursor, keys = cache.scan(
                count=SCAN_KEYS_COUNT, cursor=cursor, match='{}*'.format(
                    REDIS_LOCK_NAME_PREFIX
                )
            )
            if keys:
                cache.delete(*keys)

            if cursor == 0:
                break

    def __init__(self, name, timeout):
        if redis.VERSION < REDIS_LOCK_VERSION_REQUIRED:
            raise DependenciesException(
                'The Redis lock backend requires the Redis Python client '
                'version {} or later.'.format(
                    '.'.join(map(force_text, REDIS_LOCK_VERSION_REQUIRED))
                )
            )

        self.name = name
        redis_lock_instance = self.__class__.get_redis_connection().lock(
            name='{}{}'.format(REDIS_LOCK_NAME_PREFIX, name),
            timeout=timeout, sleep=0.1, blocking_timeout=0.1
        )
        if redis_lock_instance.acquire():
            self.redis_lock_instance = redis_lock_instance
        else:
            raise LockError

    def _release(self):
        try:
            self.redis_lock_instance.release()
        except redis.exceptions.LockNotOwnedError:
            return



from django.db import migrations


def operation_document_file_filename_copy(apps, schema_editor):
    cursor_main = schema_editor.connection.cursor()
    cursor_document_file = schema_editor.connection.cursor()
    query_document_file = '''
        UPDATE "documents_documentfile" SET "filename" = %s WHERE "documents_documentfile"."id" = %s;
    '''
    query = '''
        SELECT
            "documents_document"."label",
            "documents_documentfile"."id"
        FROM "documents_documentfile"
        INNER JOIN "documents_document" ON (
            "documents_documentfile"."document_id" = "documents_document"."id"
        )
    '''
    cursor_main.execute(query)

    for row in cursor_main.fetchall():
        cursor_document_file.execute(
            query_document_file, row
        )


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0066_documentfile_filename'),
    ]

    operations = [
        migrations.RunPython(
            code=operation_document_file_filename_copy,
            reverse_code=migrations.RunPython.noop
        ),
    ]



class DuplicatedDocumentAPIViewTestMixin:
    def _request_test_duplicated_document_list_api_view(self):
        return self.get(viewname='rest_api:duplicateddocument-list')

    def _request_test_document_duplicates_list_api_view(self):
        return self.get(
            viewname='rest_api:documentduplicate-list', kwargs={
                'document_id': self.test_documents[0].pk
            }
        )


class DuplicatedDocumentTestMixin:
    def _upload_duplicate_document(self):
        self._upload_test_document(label='duplicated document label')


class DuplicatedDocumentViewTestMixin:
    def _request_test_document_duplicates_list_view(self):
        return self.get(
            viewname='duplicates:document_duplicates_list', kwargs={
                'document_id': self.test_documents[0].pk
            }
        )

    def _request_test_duplicated_document_list_view(self):
        return self.get(viewname='duplicates:duplicated_document_list')



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link

from ..icons import (
    icon_favorite_document_add, icon_favorite_document_list,
    icon_favorite_document_remove
)
from ..permissions import permission_document_view


link_document_list_favorites = Link(
    icon=icon_favorite_document_list, text=_('Favorites'),
    view='documents:document_favorite_list'
)
link_document_favorites_add = Link(
    args='resolved_object.id', icon=icon_favorite_document_add,
    permissions=(permission_document_view,), text=_('Add to favorites'),
    view='documents:document_favorite_add'
)
link_document_favorites_remove = Link(
    args='resolved_object.id', icon=icon_favorite_document_remove,
    permissions=(permission_document_view,), text=_('Remove from favorites'),
    view='documents:document_favorite_remove'
)
link_document_multiple_favorites_add = Link(
    text=_('Add to favorites'), icon=icon_favorite_document_add,
    view='documents:document_multiple_favorite_add'
)
link_document_multiple_favorites_remove = Link(
    text=_('Remove from favorites'), icon=icon_favorite_document_remove,
    view='documents:document_multiple_favorite_remove'
)



from django.db.models import Q

from ..classes import WorkflowAction
from ..models import (
    Workflow, WorkflowRuntimeProxy, WorkflowState, WorkflowStateAction,
    WorkflowStateRuntimeProxy, WorkflowTransition, WorkflowTransitionField
)

from .literals import (
    DOCUMENT_WORKFLOW_LAUNCH_ACTION_CLASS_PATH,
    TEST_WORKFLOW_INSTANCE_LOG_ENTRY_COMMENT,
    TEST_WORKFLOW_TEMPLATE_INTERNAL_NAME,
    TEST_WORKFLOW_TEMPLATE_LABEL, TEST_WORKFLOW_TEMPLATE_LABEL_EDITED,
    TEST_WORKFLOW_TEMPLATE_STATE_ACTION_LABEL,
    TEST_WORKFLOW_TEMPLATE_STATE_ACTION_LABEL_EDITED,
    TEST_WORKFLOW_TEMPLATE_STATE_ACTION_DOTTED_PATH,
    TEST_WORKFLOW_TEMPLATE_STATE_ACTION_WHEN,
    TEST_WORKFLOW_TEMPLATE_STATE_COMPLETION,
    TEST_WORKFLOW_TEMPLATE_STATE_LABEL,
    TEST_WORKFLOW_TEMPLATE_STATE_LABEL_EDITED,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_HELP_TEXT,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_LABEL,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_LABEL_EDITED,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_NAME,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_TYPE,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL_EDITED,
)


class DocumentTypeAddRemoveWorkflowTemplateViewTestMixin:
    def _request_test_document_type_workflow_template_add_remove_get_view(self):
        return self.get(
            viewname='document_states:document_type_workflow_templates',
            kwargs={
                'document_type_id': self.test_document_type.pk,
            }
        )

    def _request_test_document_type_workflow_template_add_view(self):
        return self.post(
            viewname='document_states:document_type_workflow_templates',
            kwargs={
                'document_type_id': self.test_document_type.pk,
            }, data={
                'available-submit': 'true',
                'available-selection': self.test_workflow_template.pk
            }
        )

    def _request_test_document_type_workflow_template_remove_view(self):
        return self.post(
            viewname='document_states:document_type_workflow_templates',
            kwargs={
                'document_type_id': self.test_document_type.pk,
            }, data={
                'added-submit': 'true',
                'added-selection': self.test_workflow_template.pk
            }
        )


class WorkflowTemplateDocumentTypeViewTestMixin:
    def _request_test_workflow_template_document_type_add_remove_get_view(self):
        return self.get(
            viewname='document_states:workflow_template_document_types',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_document_type_add_view(self):
        return self.post(
            viewname='document_states:workflow_template_document_types',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
            }, data={
                'available-submit': 'true',
                'available-selection': self.test_document_type.pk
            }
        )

    def _request_test_workflow_template_document_type_remove_view(self):
        return self.post(
            viewname='document_states:workflow_template_document_types',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
            }, data={
                'added-submit': 'true',
                'added-selection': self.test_document_type.pk
            }
        )


class WorkflowInstanceAPIViewTestMixin:
    def _request_test_workflow_instance_detail_api_view(self):
        return self.get(
            viewname='rest_api:workflow-instance-detail', kwargs={
                'document_id': self.test_document.pk,
                'workflow_instance_id': self.test_document.workflows.first().pk
            }
        )

    def _request_test_workflow_instance_list_api_view(self):
        return self.get(
            viewname='rest_api:workflow-instance-list', kwargs={
                'document_id': self.test_document.pk
            }
        )

    def _request_test_workflow_instance_log_entry_create_api_view(self, workflow_instance, extra_data=None):
        data = {
            'transition_id': self.test_workflow_template_transition.pk
        }

        if extra_data:
            data.update(extra_data)

        return self.post(
            viewname='rest_api:workflow-instance-log-entry-list', kwargs={
                'document_id': self.test_document.pk,
                'workflow_instance_id': workflow_instance.pk
            }, data=data
        )

    def _request_test_workflow_instance_log_entry_list_api_view(self):
        return self.get(
            viewname='rest_api:workflow-instance-log-entry-list', kwargs={
                'document_id': self.test_document.pk,
                'workflow_instance_id': self.test_document.workflows.first().pk
            }
        )


class WorkflowTemplateLaunchActionViewTestMixin:
    def _request_document_workflow_template_launch_action_create_view(self):
        return self.post(
            viewname='document_states:workflow_template_state_action_create',
            kwargs={
                'workflow_template_state_id': self.test_workflow_template_state.pk,
                'class_path': DOCUMENT_WORKFLOW_LAUNCH_ACTION_CLASS_PATH
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_STATE_ACTION_LABEL,
                'when': TEST_WORKFLOW_TEMPLATE_STATE_ACTION_WHEN,
                'workflows': self.test_workflow_template.pk
            }
        )


class DocumentWorkflowTemplateViewTestMixin:
    def _request_test_document_single_workflow_template_launch_view(self):
        return self.post(
            viewname='document_states:document_single_workflow_templates_launch',
            kwargs={
                'document_id': self.test_document.pk
            }, data={
                'workflows': self.test_workflow_template.pk
            }
        )


class TestWorkflowAction(WorkflowAction):
    label = 'test workflow state action'

    def execute(self, context):
        context['workflow_instance']._workflow_state_action_executed = True


class WorkflowTemplateAPIViewTestMixin:
    def _request_test_workflow_template_create_api_view(self, extra_data=None):
        data = {
            'internal_name': TEST_WORKFLOW_TEMPLATE_INTERNAL_NAME,
            'label': TEST_WORKFLOW_TEMPLATE_LABEL,
        }

        if extra_data:
            data.update(extra_data)

        pk_list = list(Workflow.objects.values('pk'))

        response = self.post(
            viewname='rest_api:workflow-template-list', data=data
        )

        try:
            self.test_workflow_template = Workflow.objects.get(
                ~Q(pk__in=pk_list)
            )
        except Workflow.DoesNotExist:
            self.test_workflow_template = None

        return response

    def _request_test_workflow_template_delete_api_view(self):
        return self.delete(
            viewname='rest_api:workflow-template-detail', kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_detail_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-detail', kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_edit_via_patch_view(self):
        return self.patch(
            viewname='rest_api:workflow-template-detail', kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }, data={'label': TEST_WORKFLOW_TEMPLATE_LABEL_EDITED}
        )

    def _request_test_workflow_template_edit_via_put_view(self):
        return self.put(
            viewname='rest_api:workflow-template-detail', kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }, data={
                'internal_name': TEST_WORKFLOW_TEMPLATE_INTERNAL_NAME,
                'label': TEST_WORKFLOW_TEMPLATE_LABEL_EDITED
            }
        )

    def _request_test_workflow_template_image_view_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-image', kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_list_api_view(self):
        return self.get(viewname='rest_api:workflow-template-list')


class WorkflowInstanceViewTestMixin:
    def _request_test_document_workflow_instance_list_view(self):
        return self.get(
            viewname='document_states:workflow_instance_list', kwargs={
                'document_id': self.test_document.pk
            }
        )

    def _request_test_workflow_instance_detail_view(self):
        return self.get(
            viewname='document_states:workflow_instance_detail',
            kwargs={
                'workflow_instance_id': self.test_workflow_instance.pk
            }
        )

    def _request_test_workflow_instance_transition_execute_view(self):
        return self.post(
            viewname='document_states:workflow_instance_transition_execute',
            kwargs={
                'workflow_instance_id': self.test_workflow_instance.pk,
                'workflow_template_transition_id': self.test_workflow_template_transition.pk,
            }
        )

    def _request_test_workflow_instance_transition_selection_get_view(self):
        return self.get(
            viewname='document_states:workflow_instance_transition_selection',
            kwargs={
                'workflow_instance_id': self.test_workflow_instance.pk,
            }
        )

    def _request_test_workflow_instance_transition_selection_post_view(self):
        return self.post(
            viewname='document_states:workflow_instance_transition_selection',
            kwargs={
                'workflow_instance_id': self.test_workflow_instance.pk,
            }, data={
                'transition': self.test_workflow_template_transition.pk,
            }
        )


class WorkflowTemplateDocumentTypeAPIViewMixin:
    def _request_test_workflow_template_document_type_add_api_view(self):
        return self.post(
            viewname='rest_api:workflow-template-document-type-add',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
            }, data={
                'document_type_id': self.test_document_type.pk
            }
        )

    def _request_test_workflow_template_document_type_list_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-document-type-list', kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_document_type_remove_api_view(self):
        return self.post(
            viewname='rest_api:workflow-template-document-type-remove',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
            }, data={
                'document_type_id': self.test_document_type.pk
            }
        )


class WorkflowRuntimeProxyStateViewTestMixin:
    def _request_test_workflow_runtime_proxy_state_list_view(self):
        return self.get(
            viewname='document_states:workflow_runtime_proxy_state_list',
            kwargs={
                'workflow_runtime_proxy_id': self.test_workflow_template.pk
            }
        )

    def _request_test_workflow_runtime_proxy_state_document_list_view(self):
        return self.get(
            viewname='document_states:workflow_runtime_proxy_state_document_list',
            kwargs={
                'workflow_runtime_proxy_state_id': self.test_workflow_template_states[0].pk
            }
        )


class WorkflowRuntimeProxyViewTestMixin:
    def _request_test_workflow_runtime_proxy_list_view(self):
        return self.get(
            viewname='document_states:workflow_runtime_proxy_list'
        )

    def _request_test_workflow_runtime_proxy_document_list_view(self):
        return self.get(
            viewname='document_states:workflow_runtime_proxy_document_list',
            kwargs={
                'workflow_runtime_proxy_id': self.test_workflow_template.pk
            }
        )


class WorkflowTemplateStateActionTestMixin:
    TestWorkflowAction = TestWorkflowAction
    test_workflow_template_state_action_path = 'mayan.apps.document_states.tests.mixins.TestWorkflowAction'

    def setUp(self):
        super().setUp()
        self.test_workflow_template_state_actions = []

    def _create_test_workflow_template_state_action(
        self, extra_data=None, workflow_state_index=0
    ):
        total_test_workflow_template_state_actions = len(
            self.test_workflow_template_state_actions
        )
        label = '{}_{}'.format(
            TEST_WORKFLOW_TEMPLATE_STATE_ACTION_LABEL,
            total_test_workflow_template_state_actions
        )

        data = {
            'label': label,
            'action_path': self.test_workflow_template_state_action_path
        }

        if extra_data:
            data.update(extra_data)

        self.test_workflow_template_state_action = self.test_workflow_template_states[
            workflow_state_index
        ].actions.create(
            **data
        )

        self.test_workflow_template_state_actions.append(
            self.test_workflow_template_state_action
        )


class WorkflowTemplateStateActionViewTestMixin:
    def _request_test_workflow_template_state_action_create_get_view(self, class_path):
        return self.get(
            viewname='document_states:workflow_template_state_action_create',
            kwargs={
                'workflow_template_state_id': self.test_workflow_template_state.pk,
                'class_path': class_path
            }
        )

    def _request_test_workflow_template_state_action_create_post_view(
        self, class_path, extra_data=None
    ):
        data = {
            'label': TEST_WORKFLOW_TEMPLATE_STATE_ACTION_LABEL,
            'when': TEST_WORKFLOW_TEMPLATE_STATE_ACTION_WHEN
        }
        if extra_data:
            data.update(extra_data)

        pk_list = list(
            WorkflowStateAction.objects.values_list('pk', flat=True)
        )

        response = self.post(
            viewname='document_states:workflow_template_state_action_create',
            kwargs={
                'workflow_template_state_id': self.test_workflow_template_state.pk,
                'class_path': class_path
            }, data=data
        )

        try:
            self.test_workflow_template_state_action = WorkflowStateAction.objects.get(
                ~Q(pk__in=pk_list)
            )
        except WorkflowStateAction.DoesNotExist:
            self.test_workflow_template_state_action = None

        return response

    def _request_test_worflow_template_state_action_delete_view(self):
        return self.post(
            viewname='document_states:workflow_template_state_action_delete',
            kwargs={
                'workflow_template_state_action_id': self.test_workflow_template_state_action.pk
            }
        )

    def _request_test_worflow_template_state_action_edit_view(self):
        return self.post(
            viewname='document_states:workflow_template_state_action_edit',
            kwargs={
                'workflow_template_state_action_id': self.test_workflow_template_state_action.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_STATE_ACTION_LABEL_EDITED,
                'when': self.test_workflow_template_state_action.when
            }
        )

    def _request_test_worflow_template_state_action_list_view(self):
        return self.get(
            viewname='document_states:workflow_template_state_action_list',
            kwargs={
                'workflow_template_state_id': self.test_workflow_template_state.pk
            }
        )

    def _request_test_workflow_state_action_selection_view(self):
        return self.post(
            viewname='document_states:workflow_template_state_action_selection',
            kwargs={
                'workflow_template_state_id': self.test_workflow_template_state.pk
            }, data={
                'klass': TEST_WORKFLOW_TEMPLATE_STATE_ACTION_DOTTED_PATH
            }
        )


class WorkflowTemplateStateAPIViewTestMixin:
    def _request_test_workflow_state_create_api_view(self):
        pk_list = list(WorkflowState.objects.values('pk'))

        response = self.post(
            viewname='rest_api:workflow-template-state-list',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }, data={
                'completion': TEST_WORKFLOW_TEMPLATE_STATE_COMPLETION,
                'label': TEST_WORKFLOW_TEMPLATE_STATE_LABEL
            }
        )

        try:
            self.test_workflow_template_state = WorkflowState.objects.get(
                ~Q(pk__in=pk_list)
            )
        except WorkflowState.DoesNotExist:
            self.test_workflow_template_state = None

        return response

    def _request_test_workflow_state_delete_api_view(self):
        return self.delete(
            viewname='rest_api:workflow-template-state-detail',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
                'workflow_template_state_id': self.test_workflow_template_state.pk
            }
        )

    def _request_test_workflow_state_detail_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-state-detail',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
                'workflow_template_state_id': self.test_workflow_template_state.pk
            }
        )

    def _request_test_workflow_state_list_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-state-list', kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }
        )

    def _request_test_workflow_state_edit_patch_api_view(self):
        return self.patch(
            viewname='rest_api:workflow-template-state-detail',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
                'workflow_template_state_id': self.test_workflow_template_state.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_STATE_LABEL_EDITED
            }
        )

    def _request_test_workflow_state_edit_put_api_view(self):
        return self.put(
            viewname='rest_api:workflow-template-state-detail',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
                'workflow_template_state_id': self.test_workflow_template_state.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_STATE_LABEL_EDITED
            }
        )


class WorkflowStateViewTestMixin:
    def _request_test_workflow_state_create_view(self, extra_data=None):
        pk_list = list(WorkflowState.objects.values_list('pk', flat=True))

        data = {
            'label': TEST_WORKFLOW_TEMPLATE_STATE_LABEL,
            'completion': TEST_WORKFLOW_TEMPLATE_STATE_COMPLETION,
        }
        if extra_data:
            data.update(extra_data)

        response = self.post(
            viewname='document_states:workflow_template_state_create',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }, data=data
        )

        try:
            self.test_workflow_template_state = WorkflowState.objects.get(
                ~Q(pk__in=pk_list)
            )
        except WorkflowState.DoesNotExist:
            self.test_workflow_template_state = None

        return response

    def _request_test_workflow_state_delete_view(self):
        return self.post(
            viewname='document_states:workflow_template_state_delete',
            kwargs={
                'workflow_template_state_id': self.test_workflow_template_states[0].pk
            }
        )

    def _request_test_workflow_state_edit_view(self):
        return self.post(
            viewname='document_states:workflow_template_state_edit',
            kwargs={
                'workflow_template_state_id': self.test_workflow_template_states[0].pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_STATE_LABEL_EDITED
            }
        )

    def _request_test_workflow_state_list_view(self):
        return self.get(
            viewname='document_states:workflow_template_state_list',
            kwargs={'workflow_template_id': self.test_workflow_template.pk}
        )


class WorkflowTemplateTestMixin:
    def setUp(self):
        super().setUp()
        self.test_workflow_runtime_proxies = []
        self.test_workflow_template_state_runtime_proxies = []
        self.test_workflow_template_states = []
        self.test_workflow_template_transitions = []
        self.test_workflow_templates = []

    def _create_test_workflow_template(
        self, add_test_document_type=False, auto_launch=True
    ):
        total_test_workflow_templates = len(self.test_workflow_templates)
        label = '{}_{}'.format(
            TEST_WORKFLOW_TEMPLATE_LABEL, total_test_workflow_templates
        )
        internal_name = '{}_{}'.format(
            TEST_WORKFLOW_TEMPLATE_INTERNAL_NAME,
            total_test_workflow_templates
        )

        self.test_workflow_template = Workflow.objects.create(
            auto_launch=auto_launch, label=label,
            internal_name=internal_name
        )
        self.test_workflow_templates.append(
            self.test_workflow_template
        )
        self.test_workflow_runtime_proxy = WorkflowRuntimeProxy.objects.get(
            pk=self.test_workflow_template.pk
        )
        self.test_workflow_runtime_proxies.append(
            self.test_workflow_runtime_proxy
        )

        if add_test_document_type:
            self.test_workflow_template.document_types.add(
                self.test_document_type
            )

    def _create_test_workflow_template_state(self):
        total_test_workflow_template_states = len(
            self.test_workflow_template_states
        )
        label = '{}_{}'.format(
            TEST_WORKFLOW_TEMPLATE_STATE_LABEL,
            total_test_workflow_template_states
        )
        initial = total_test_workflow_template_states == 0

        self.test_workflow_template_state = self.test_workflow_template.states.create(
            completion=TEST_WORKFLOW_TEMPLATE_STATE_COMPLETION, initial=initial,
            label=label
        )
        self.test_workflow_template_states.append(
            self.test_workflow_template_state
        )
        self.test_workflow_template_state_runtime_proxy = WorkflowStateRuntimeProxy.objects.get(
            pk=self.test_workflow_template_state.pk
        )
        self.test_workflow_template_state_runtime_proxies.append(
            self.test_workflow_template_state_runtime_proxy
        )

    def _create_test_workflow_template_transition(self):
        total_test_workflow_template_transitions = len(
            self.test_workflow_template_transitions
        )
        label = '{}_{}'.format(
            TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL,
            total_test_workflow_template_transitions
        )

        self.test_workflow_template_transition = self.test_workflow_template.transitions.create(
            label=label,
            origin_state=self.test_workflow_template_states[0],
            destination_state=self.test_workflow_template_states[1],
        )

        self.test_workflow_template_transitions.append(
            self.test_workflow_template_transition
        )

    def _create_test_workflow_template_instance_log_entry(self):
        self.test_document.workflows.first().log_entries.create(
            comment=TEST_WORKFLOW_INSTANCE_LOG_ENTRY_COMMENT,
            transition=self.test_workflow_template_transition,
            user=self._test_case_user
        )

    def _transition_test_workflow_instance(self, extra_data=None):
        self.test_document.workflows.first().do_transition(
            comment=TEST_WORKFLOW_INSTANCE_LOG_ENTRY_COMMENT,
            extra_data=extra_data,
            transition=self.test_workflow_template_transition
        )


class WorkflowToolViewTestMixin:
    def _request_workflow_launch_view(self):
        return self.post(
            viewname='document_states:tool_launch_workflows',
        )


class WorkflowTransitionEventViewTestMixin:
    def _request_test_workflow_template_transition_event_list_view(self):
        return self.get(
            viewname='document_states:workflow_template_transition_events',
            kwargs={
                'workflow_template_transition_id': self.test_workflow_template_transition.pk
            }
        )


class WorkflowTransitionFieldViewTestMixin:
    def _request_workflow_template_transition_field_create_view(self):
        pk_list = list(
            WorkflowTransitionField.objects.values_list('pk', flat=True)
        )

        response = self.post(
            viewname='document_states:workflow_template_transition_field_create',
            kwargs={
                'workflow_template_transition_id': self.test_workflow_template_transition.pk
            }, data={
                'field_type': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_TYPE,
                'name': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_NAME,
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_LABEL,
                'help_text': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_HELP_TEXT
            }
        )

        try:
            self.test_workflow_template_transition_field = WorkflowTransitionField.objects.get(
                ~Q(pk__in=pk_list)
            )
        except WorkflowTransitionField.DoesNotExist:
            self.test_workflow_template_transition_field = None

        return response

    def _request_workflow_template_transition_field_delete_view(self):
        return self.post(
            viewname='document_states:workflow_template_transition_field_delete',
            kwargs={
                'workflow_template_transition_field_id': self.test_workflow_template_transition_field.pk
            },
        )

    def _request_workflow_template_transition_field_edit_view(self):
        return self.post(
            viewname='document_states:workflow_template_transition_field_edit',
            kwargs={
                'workflow_template_transition_field_id': self.test_workflow_template_transition_field.pk
            }, data={
                'field_type': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_TYPE,
                'name': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_NAME,
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_LABEL_EDITED,
                'help_text': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_HELP_TEXT
            }
        )

    def _request_test_workflow_template_transition_field_list_view(self):
        return self.get(
            viewname='document_states:workflow_template_transition_field_list',
            kwargs={
                'workflow_template_transition_id': self.test_workflow_template_transition.pk
            }
        )


class WorkflowTemplateTransitionAPIViewTestMixin:
    def _request_test_workflow_template_transition_create_api_view(
        self, extra_data=None
    ):
        data = {
            'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL,
            'origin_state_id': self.test_workflow_template_states[0].pk,
            'destination_state_id': self.test_workflow_template_states[1].pk,
        }

        if extra_data:
            data.update(extra_data)

        pk_list = list(
            WorkflowTransition.objects.values_list('pk', flat=True)
        )

        response = self.post(
            viewname='rest_api:workflow-template-transition-list', kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }, data=data
        )

        try:
            self.test_workflow_template_transition = WorkflowTransition.objects.get(
                ~Q(pk__in=pk_list)
            )
        except WorkflowTransition.DoesNotExist:
            self.test_workflow_template_transition = None

        return response

    def _request_test_workflow_template_transition_delete_api_view(self):
        return self.delete(
            viewname='rest_api:workflow-template-transition-detail',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
                'workflow_template_transition_id': self.test_workflow_template_transition.pk
            }
        )

    def _request_test_workflow_template_transition_detail_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-transition-detail',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
                'workflow_template_transition_id': self.test_workflow_template_transition.pk
            }
        )

    def _request_test_workflow_template_transition_list_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-transition-list',
            kwargs={'workflow_template_id': self.test_workflow_template.pk}
        )

    def _request_test_workflow_template_transition_edit_patch_api_view(self):
        return self.patch(
            viewname='rest_api:workflow-template-transition-detail',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
                'workflow_template_transition_id': self.test_workflow_template_transition.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL_EDITED,
                'origin_state_id': self.test_workflow_template_states[1].pk,
                'destination_state_id': self.test_workflow_template_states[0].pk,
            }
        )

    def _request_test_workflow_template_transition_edit_put_api_view_via(self):
        return self.put(
            viewname='rest_api:workflow-template-transition-detail',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
                'workflow_template_transition_id': self.test_workflow_template_transition.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL_EDITED,
                'origin_state_id': self.test_workflow_template_states[1].pk,
                'destination_state_id': self.test_workflow_template_states[0].pk,
            }
        )


class WorkflowTransitionFieldAPIViewTestMixin:
    def _request_test_workflow_template_transition_field_create_api_view(self):
        pk_list = list(WorkflowTransitionField.objects.values_list('pk'))

        response = self.post(
            viewname='rest_api:workflow-template-transition-field-list',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
                'workflow_template_transition_id': self.test_workflow_template_transition.pk,
            }, data={
                'field_type': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_TYPE,
                'name': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_NAME,
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_LABEL,
                'help_text': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_HELP_TEXT
            }
        )

        try:
            self.test_workflow_template_transition_field = WorkflowTransitionField.objects.get(
                ~Q(pk__in=pk_list)
            )
        except WorkflowTransitionField.DoesNotExist:
            self.test_workflow_template_transition_field = None

        return response

    def _request_test_workflow_template_transition_field_delete_api_view(self):
        return self.delete(
            viewname='rest_api:workflow-template-transition-field-detail',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
                'workflow_template_transition_id': self.test_workflow_template_transition.pk,
                'workflow_template_transition_field_id': self.test_workflow_template_transition_field.pk,
            }
        )

    def _request_test_workflow_template_transition_field_detail_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-transition-field-detail',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
                'workflow_template_transition_id': self.test_workflow_template_transition.pk,
                'workflow_template_transition_field_id': self.test_workflow_template_transition_field.pk,
            }
        )

    def _request_test_workflow_template_transition_field_edit_via_patch_api_view(self):
        return self.patch(
            viewname='rest_api:workflow-template-transition-field-detail',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
                'workflow_template_transition_id': self.test_workflow_template_transition.pk,
                'workflow_template_transition_field_id': self.test_workflow_template_transition_field.pk,
            }, data={
                'label': '{} edited'.format(
                    self.test_workflow_template_transition_field
                )
            }
        )

    def _request_test_workflow_template_transition_field_list_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-transition-field-list',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
                'workflow_template_transition_id': self.test_workflow_template_transition.pk,
            }
        )


class WorkflowTransitionFieldTestMixin:
    def _create_test_workflow_template_transition_field(self, extra_data=None):
        kwargs = {
            'field_type': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_TYPE,
            'name': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_NAME,
            'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_LABEL,
            'help_text': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_HELP_TEXT
        }
        kwargs.update(extra_data or {})

        self.test_workflow_template_transition_field = self.test_workflow_template_transition.fields.create(
            **kwargs
        )


class WorkflowTransitionViewTestMixin:
    def _request_test_workflow_template_transition_create_view(self):
        pk_list = list(WorkflowTransition.objects.values_list('pk', flat=True))

        response = self.post(
            viewname='document_states:workflow_template_transition_create',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL,
                'origin_state': self.test_workflow_template_states[0].pk,
                'destination_state': self.test_workflow_template_states[1].pk,
            }
        )

        try:
            self.test_workflow_template_transition = WorkflowTransition.objects.get(
                ~Q(pk__in=pk_list)
            )
        except WorkflowTransition.DoesNotExist:
            self.test_workflow_template_transition = None

        return response

    def _request_test_workflow_template_transition_delete_view(self):
        return self.post(
            viewname='document_states:workflow_template_transition_delete',
            kwargs={
                'workflow_template_transition_id': self.test_workflow_template_transition.pk
            }
        )

    def _request_test_workflow_template_transition_edit_view(self):
        return self.post(
            viewname='document_states:workflow_template_transition_edit',
            kwargs={
                'workflow_template_transition_id': self.test_workflow_template_transition.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL_EDITED,
                'origin_state': self.test_workflow_template_states[0].pk,
                'destination_state': self.test_workflow_template_states[1].pk,
            }
        )

    def _request_test_workflow_template_transition_list_view(self):
        return self.get(
            viewname='document_states:workflow_template_transition_list',
            kwargs={'workflow_template_id': self.test_workflow_template.pk}
        )


class WorkflowTemplateViewTestMixin:
    def _request_test_workflow_template_create_view(self):
        data = {
            'internal_name': TEST_WORKFLOW_TEMPLATE_INTERNAL_NAME,
            'label': TEST_WORKFLOW_TEMPLATE_LABEL,
        }

        pk_list = list(Workflow.objects.values('pk'))

        response = self.post(
            viewname='document_states:workflow_template_create', data=data
        )

        try:
            self.test_workflow_template = Workflow.objects.get(
                ~Q(pk__in=pk_list)
            )
        except Workflow.DoesNotExist:
            self.test_workflow_template = None

        return response

    def _request_test_workflow_template_delete_view(self):
        return self.post(
            viewname='document_states:workflow_template_single_delete',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_edit_view(self):
        return self.post(
            viewname='document_states:workflow_template_edit', kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
            }, data={
                'auto_launch': True,
                'label': TEST_WORKFLOW_TEMPLATE_LABEL_EDITED,
                'internal_name': self.test_workflow_template.internal_name
            }
        )

    def _request_test_workflow_template_launch_view(self):
        return self.post(
            viewname='document_states:workflow_template_launch', kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_list_view(self):
        return self.get(
            viewname='document_states:workflow_template_list',
        )

    def _request_test_workflow_template_preview_view(self):
        return self.get(
            viewname='document_states:workflow_template_preview', kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
            }
        )



from django.apps import apps

from mayan.celery import app


@app.task(ignore_result=True)
def task_duplicates_clean_empty_lists():
    DuplicateBackendEntry = apps.get_model(
        app_label='duplicates', model_name='DuplicateBackendEntry'
    )
    DuplicateBackendEntry.objects.clean_empty_duplicate_lists()


@app.task(ignore_result=True)
def task_duplicates_scan_all():
    StoredDuplicateBackend = apps.get_model(
        app_label='duplicates', model_name='StoredDuplicateBackend'
    )

    StoredDuplicateBackend.objects.scan_all()


@app.task(ignore_result=True)
def task_duplicates_scan_for(document_id):
    Document = apps.get_model(
        app_label='documents', model_name='Document'
    )
    StoredDuplicateBackend = apps.get_model(
        app_label='duplicates', model_name='StoredDuplicateBackend'
    )

    document = Document.objects.get(pk=document_id)

    StoredDuplicateBackend.objects.scan_document(document=document)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0056_auto_20200916_0959'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='DocumentVersion',
            new_name='DocumentFile',
        ),
        migrations.RunSQL
        (
            sql='DROP INDEX "documents_documentversion_timestamp_30bada95";',
            reverse_sql='CREATE INDEX "documents_documentversion_timestamp_30bada95" '
            'ON "documents_documentversion" ("timestamp");',
        ),
        migrations.RunSQL
        (
            sql='DROP INDEX "documents_documentversion_document_id_42757b7a";',
            reverse_sql='CREATE INDEX "documents_documentversion_document_id_42757b7a" '
            'ON "documents_documentversion" ("document_id");',
        ),
        migrations.AlterModelOptions(
            name='documentpageresult',
            options={
                'ordering': (
                    'document_file__document', 'page_number'
                ),
                'verbose_name': 'Document page',
                'verbose_name_plural': 'Document pages'
            },
        ),
        migrations.RenameField(
            model_name='documentpage',
            old_name='document_version',
            new_name='document_file',
        ),
    ]



from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from .mixins import (
    DuplicatedDocumentTestMixin, DuplicatedDocumentViewTestMixin
)


class DocumentsDuplicateListViewsTestCase(
    DuplicatedDocumentTestMixin, DuplicatedDocumentViewTestMixin,
    GenericDocumentViewTestCase
):
    def test_document_duplicates_list_no_permission(self):
        self._upload_duplicate_document()

        response = self._request_test_document_duplicates_list_view()
        self.assertEqual(response.status_code, 404)

    def test_document_duplicates_list_with_source_access(self):
        self._upload_duplicate_document()
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_document_view
        )

        response = self._request_test_document_duplicates_list_view()
        self.assertContains(
            response=response, status_code=200,
            text=self.test_documents[0].label
        )
        self.assertNotContains(
            response=response, status_code=200,
            text=self.test_documents[1].label
        )

    def test_document_duplicates_list_with_target_access(self):
        self._upload_duplicate_document()

        self.grant_access(
            obj=self.test_documents[1],
            permission=permission_document_view
        )

        response = self._request_test_document_duplicates_list_view()
        self.assertEqual(response.status_code, 404)

    def test_document_duplicates_list_with_full_access(self):
        self._upload_duplicate_document()
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_documents[1],
            permission=permission_document_view
        )

        response = self._request_test_document_duplicates_list_view()
        self.assertContains(
            response=response, status_code=200,
            text=self.test_documents[0].label
        )
        self.assertContains(
            response=response, status_code=200,
            text=self.test_documents[1].label
        )

    def test_document_duplicates_list_trashed_source_with_full_access(self):
        self._upload_duplicate_document()
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_documents[1],
            permission=permission_document_view
        )

        self.test_documents[0].delete()

        response = self._request_test_document_duplicates_list_view()
        self.assertEqual(response.status_code, 404)

    def test_document_duplicates_list_trashed_target_with_full_access(self):
        self._upload_duplicate_document()
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_documents[1],
            permission=permission_document_view
        )

        self.test_documents[1].delete()

        response = self._request_test_document_duplicates_list_view()
        self.assertContains(
            response=response, status_code=200,
            text=self.test_documents[0].label
        )
        self.assertNotContains(
            response=response, status_code=200,
            text=self.test_documents[1].label
        )


class DuplicatedDocumentListViewsTestCase(
    DuplicatedDocumentTestMixin, DuplicatedDocumentViewTestMixin,
    GenericDocumentViewTestCase
):
    def test_duplicated_document_list_no_permission(self):
        self._upload_duplicate_document()

        response = self._request_test_duplicated_document_list_view()
        self.assertNotContains(
            response=response, status_code=200,
            text=self.test_documents[0].label
        )

    def test_duplicated_document_list_with_source_access(self):
        self._upload_duplicate_document()
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_document_view
        )

        response = self._request_test_duplicated_document_list_view()
        self.assertNotContains(
            response=response, status_code=200,
            text=self.test_documents[0].label
        )
        self.assertNotContains(
            response=response, status_code=200,
            text=self.test_documents[1].label
        )

    def test_duplicated_document_list_with_target_access(self):
        self._upload_duplicate_document()
        self.grant_access(
            obj=self.test_documents[1],
            permission=permission_document_view
        )

        response = self._request_test_duplicated_document_list_view()
        self.assertNotContains(
            response=response, status_code=200,
            text=self.test_documents[0].label
        )
        self.assertNotContains(
            response=response, status_code=200,
            text=self.test_documents[1].label
        )

    def test_duplicated_document_list_with_full_access(self):
        self._upload_duplicate_document()
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_documents[1],
            permission=permission_document_view
        )

        response = self._request_test_duplicated_document_list_view()
        self.assertContains(
            response=response, status_code=200,
            text=self.test_documents[0].label
        )
        self.assertContains(
            response=response, status_code=200,
            text=self.test_documents[1].label
        )

    def test_duplicated_document_list_trashed_source_with_full_access(self):
        self._upload_duplicate_document()
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_documents[1],
            permission=permission_document_view
        )

        self.test_documents[0].delete()

        response = self._request_test_duplicated_document_list_view()
        self.assertNotContains(
            response=response, status_code=200,
            text=self.test_documents[0].label
        )
        self.assertNotContains(
            response=response, status_code=200,
            text=self.test_documents[1].label
        )

    def test_duplicated_document_list_trashed_target_with_full_access(self):
        self._upload_duplicate_document()
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_documents[1],
            permission=permission_document_view
        )

        self.test_documents[1].delete()

        response = self._request_test_duplicated_document_list_view()
        self.assertNotContains(
            response=response, status_code=200,
            text=self.test_documents[0].label
        )
        self.assertNotContains(
            response=response, status_code=200,
            text=self.test_documents[1].label
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.converter.permissions import (
    permission_transformation_delete, permission_transformation_edit
)
from mayan.apps.navigation.classes import Link

from ..icons import (
    icon_document_file_delete, icon_document_file_download_quick,
    icon_document_file_edit, icon_document_file_list,
    icon_document_file_print, icon_document_file_properties,
    icon_document_file_return_to_document, icon_document_file_return_list,
    icon_document_file_preview, icon_document_file_transformations_clear,
    icon_document_file_transformations_clone
)
from ..permissions import (
    permission_document_file_delete, permission_document_file_download,
    permission_document_file_edit, permission_document_file_print,
    permission_document_file_view, permission_document_view
)

link_document_file_delete = Link(
    args='object.pk',
    icon=icon_document_file_delete,
    permissions=(permission_document_file_delete,), tags='dangerous',
    text=_('Delete'), view='documents:document_file_delete',
)
link_document_file_delete_multiple = Link(
    icon=icon_document_file_delete,
    permissions=(permission_document_file_delete,), tags='dangerous',
    text=_('Delete'), view='documents:document_file_delete_multiple',
)
link_document_file_edit = Link(
    args='object.pk', icon=icon_document_file_edit,
    permissions=(permission_document_file_edit,),
    text=_('Edit'), view='documents:document_file_edit',
)
link_document_file_download_quick = Link(
    args='resolved_object.id', icon=icon_document_file_download_quick,
    permissions=(permission_document_file_download,),
    text=_('Quick download'), view='documents:document_file_download'
)
link_document_file_list = Link(
    args='resolved_object.pk',
    icon=icon_document_file_list,
    permissions=(permission_document_file_view,),
    text=_('Files'), view='documents:document_file_list',
)
link_document_file_print_form = Link(
    args='resolved_object.id', icon=icon_document_file_print,
    permissions=(permission_document_file_print,), text=_('Print'),
    view='documents:document_file_print_form'
)
link_document_file_properties = Link(
    args='resolved_object.id',
    icon=icon_document_file_properties,
    permissions=(permission_document_file_view,),
    text=_('Properties'), view='documents:document_file_properties',
)
link_document_file_return_to_document = Link(
    args='resolved_object.document.pk',
    icon=icon_document_file_return_to_document,
    permissions=(permission_document_view,), text=_('Document'),
    view='documents:document_preview',
)
link_document_file_return_list = Link(
    args='resolved_object.document.pk',
    icon=icon_document_file_return_list,
    permissions=(permission_document_file_view,), text=_('Files'),
    view='documents:document_file_list',
)
link_document_file_preview = Link(
    args='resolved_object.pk',
    icon=icon_document_file_preview,
    permissions=(permission_document_file_view,),
    text=_('Preview'), view='documents:document_file_preview'
)
link_document_file_transformations_clear = Link(
    args='resolved_object.id',
    icon=icon_document_file_transformations_clear,
    permissions=(permission_transformation_delete,),
    text=_('Clear transformations'),
    view='documents:document_file_transformations_clear'
)
link_document_file_multiple_transformations_clear = Link(
    icon=icon_document_file_transformations_clear,
    permissions=(permission_transformation_delete,),
    text=_('Clear transformations'),
    view='documents:document_file_multiple_transformations_clear'
)
link_document_file_transformations_clone = Link(
    args='resolved_object.id',
    icon=icon_document_file_transformations_clone,
    permissions=(permission_transformation_edit,),
    text=_('Clone transformations'),
    view='documents:document_file_transformations_clone'
)



from datetime import timedelta

from django.utils.timezone import now

from mayan.apps.common.settings import settings_db_sync_task_delay

from .events import event_ocr_document_version_submit
from .tasks import task_document_version_ocr_process


def method_document_ocr_submit(self):
    version_active = self.version_active
    # Don't error out if document has no version
    if version_active:
        version_active.submit_for_ocr()


def method_document_version_ocr_submit(self):
    event_ocr_document_version_submit.commit(
        action_object=self.document, target=self
    )

    task_document_version_ocr_process.apply_async(
        eta=now() + timedelta(seconds=settings_db_sync_task_delay.value),
        kwargs={'document_version_id': self.pk}
    )



import logging

from celery import chain, chord

from django.apps import apps
from django.db import OperationalError

from mayan.apps.documents.tasks import task_document_version_page_image_generate
from mayan.celery import app

from .events import event_ocr_document_version_finish
from .literals import DO_OCR_RETRY_DELAY
from .signals import signal_post_document_version_ocr

logger = logging.getLogger(name=__name__)


@app.task(bind=True, default_retry_delay=DO_OCR_RETRY_DELAY, ignore_result=True)
def task_document_version_ocr_process(self, document_version_id):
    logger.info(
        'Starting OCR for document file page: %s', document_version_id
    )
    DocumentVersion = apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )

    document_version = DocumentVersion.objects.get(
        pk=document_version_id
    )

    try:
        page_tasks = []
        for document_version_page in document_version.pages.all():
            page_tasks.append(
                chain(
                    task_document_version_page_image_generate.s(
                        document_version_page_id=document_version_page.pk
                    ),
                    task_document_version_page_ocr_process.s(
                        document_version_page_id=document_version_page.pk
                    )
                )
            )
        chord(page_tasks)(
            task_document_version_ocr_finished.s(
                document_version_id=document_version.pk
            )
        )
    except Exception as exception:
        document_version.ocr_errors.create(result=exception)
        raise
    except OperationalError as exception:
        logger.warning(
            'OCR operational error for document file: %d; %s. Retrying.',
            document_version_id, exception
        )
        raise self.retry(exc=exception)


@app.task(bind=True, default_retry_delay=DO_OCR_RETRY_DELAY, ignore_result=True)
def task_document_version_page_ocr_process(
    self, cache_filename, document_version_page_id
):
    DocumentVersionPageOCRContent = apps.get_model(
        app_label='ocr', model_name='DocumentVersionPageOCRContent'
    )
    DocumentVersionPage = apps.get_model(
        app_label='documents', model_name='DocumentVersionPage'
    )
    document_version_page = DocumentVersionPage.objects.get(
        pk=document_version_page_id
    )

    try:
        DocumentVersionPageOCRContent.objects.process_document_version_page(
            cache_filename=cache_filename,
            document_version_page=document_version_page,
        )
    except OperationalError as exception:
        raise self.retry(exc=exception)


@app.task(bind=True, ignore_result=True)
def task_document_version_ocr_finished(self, results, document_version_id):
    logger.info(
        'OCR complete for document file: %s', document_version_id
    )

    DocumentVersion = apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )
    document_version = DocumentVersion.objects.get(pk=document_version_id)

    document_version.ocr_errors.all().delete()

    try:
        event_ocr_document_version_finish.commit(
            action_object=document_version.document,
            target=document_version
        )

        signal_post_document_version_ocr.send(
            sender=document_version.__class__,
            instance=document_version
        )
    except Exception as exception:
        logger.error(
            'Exception in OCR finish for document file: %d; %s',
            document_version_id,
            exception
        )
        document_version.ocr_errors.create(result=exception)
        raise
    except OperationalError as exception:
        logger.warning(
            'Operational error in OCR finish for document file: %d; %s. '
            'Retrying.', document_version, exception
        )
        raise self.retry(exc=exception)



import logging

from django.apps import apps
from django.db import models, transaction

from .classes import OCRBackendBase
from .events import event_ocr_document_version_content_deleted

logger = logging.getLogger(name=__name__)


class DocumentVersionPageOCRContentManager(models.Manager):
    def delete_content_for(self, document_version, user=None):
        with transaction.atomic():
            for document_version_page in document_version.pages.all():
                self.filter(
                    document_version_page=document_version_page
                ).delete()

            event_ocr_document_version_content_deleted.commit(
                actor=user, action_object=document_version.document,
                target=document_version
            )

    def process_document_version_page(
        self, cache_filename, document_version_page
    ):
        logger.info(
            'Processing page: %d of document version: %s',
            document_version_page.page_number,
            document_version_page.document_version
        )

        DocumentVersionPageOCRContent = apps.get_model(
            app_label='ocr', model_name='DocumentVersionPageOCRContent'
        )

        try:
            with document_version_page.cache_partition.get_file(filename=cache_filename).open() as file_object:
                ocr_content = OCRBackendBase.get_instance().execute(
                    file_object=file_object,
                    language=document_version_page.document_version.document.language
                )
                DocumentVersionPageOCRContent.objects.update_or_create(
                    document_version_page=document_version_page, defaults={
                        'content': ocr_content
                    }
                )
        except Exception as exception:
            logger.error(
                'OCR error for document page: %d; %s',
                document_version_page.pk,
                exception
            )
            raise
        else:
            logger.info(
                'Finished processing page: %d of document version: %s',
                document_version_page.page_number,
                document_version_page.document_version
            )


class DocumentTypeSettingsManager(models.Manager):
    def get_by_natural_key(self, document_type_natural_key):
        DocumentType = apps.get_model(
            app_label='documents', model_name='DocumentType'
        )
        try:
            document_type = DocumentType.objects.get_by_natural_key(
                document_type_natural_key
            )
        except DocumentType.DoesNotExist:
            raise self.model.DoesNotExist

        return self.get(document_type__pk=document_type.pk)



import logging

from django.apps import apps
from django.db import models
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.common.literals import TIME_DELTA_UNIT_CHOICES
from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.common.serialization import yaml_load
from mayan.apps.common.validators import YAMLValidator
from mayan.apps.events.classes import (
    EventManagerMethodAfter, EventManagerSave
)
from mayan.apps.events.decorators import method_event

from ..classes import BaseDocumentFilenameGenerator
from ..events import (
    event_document_type_created, event_document_type_edited,
    event_document_type_quick_label_created,
    event_document_type_quick_label_deleted,
    event_document_type_quick_label_edited
)
from ..literals import DEFAULT_DELETE_PERIOD, DEFAULT_DELETE_TIME_UNIT
from ..managers import DocumentTypeManager
from ..permissions import permission_document_view
from ..settings import setting_language

__all__ = ('DocumentType', 'DocumentTypeFilename')
logger = logging.getLogger(name=__name__)


class DocumentType(ExtraDataModelMixin, models.Model):
    """
    Define document types or classes to which a specific set of
    properties can be attached
    """
    label = models.CharField(
        help_text=_('The name of the document type.'), max_length=96,
        unique=True, verbose_name=_('Label')
    )
    trash_time_period = models.PositiveIntegerField(
        blank=True, help_text=_(
            'Amount of time after which documents of this type will be '
            'moved to the trash.'
        ), null=True, verbose_name=_('Trash time period')
    )
    trash_time_unit = models.CharField(
        blank=True, choices=TIME_DELTA_UNIT_CHOICES, null=True, max_length=8,
        verbose_name=_('Trash time unit')
    )
    delete_time_period = models.PositiveIntegerField(
        blank=True, default=DEFAULT_DELETE_PERIOD, help_text=_(
            'Amount of time after which documents of this type in the trash '
            'will be deleted.'
        ), null=True, verbose_name=_('Delete time period')
    )
    delete_time_unit = models.CharField(
        blank=True, choices=TIME_DELTA_UNIT_CHOICES,
        default=DEFAULT_DELETE_TIME_UNIT, max_length=8, null=True,
        verbose_name=_('Delete time unit')
    )
    filename_generator_backend = models.CharField(
        default=BaseDocumentFilenameGenerator.get_default(), help_text=_(
            'The class responsible for producing the actual filename used '
            'to store the uploaded documents.'
        ), max_length=224, verbose_name=_('Filename generator backend')
    )
    filename_generator_backend_arguments = models.TextField(
        blank=True, help_text=_(
            'The arguments for the filename generator backend as a '
            'YAML dictionary.'
        ), validators=[YAMLValidator()], verbose_name=_(
            'Filename generator backend arguments'
        )
    )

    objects = DocumentTypeManager()

    class Meta:
        ordering = ('label',)
        verbose_name = _('Document type')
        verbose_name_plural = _('Documents types')

    def __str__(self):
        return self.label

    def delete(self, *args, **kwargs):
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )

        for document in Document.objects.filter(document_type=self):
            document.delete(to_trash=False)

        return super().delete(*args, **kwargs)

    @property
    def trashed_documents(self):
        TrashedDocument = apps.get_model(
            app_label='documents', model_name='TrashedDocument'
        )

        return TrashedDocument.objects.filter(document_type=self)

    def get_absolute_url(self):
        return reverse(
            viewname='documents:document_type_document_list', kwargs={
                'document_type_id': self.pk
            }
        )

    def get_document_count(self, user):
        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_document_view, queryset=self.documents,
            user=user
        )

        return queryset.count()

    def get_upload_filename(self, instance, filename):
        generator_klass = BaseDocumentFilenameGenerator.get(
            name=self.filename_generator_backend
        )
        generator_instance = generator_klass(
            **yaml_load(
                stream=self.filename_generator_backend_arguments or '{}'
            )
        )
        return generator_instance.upload_to(
            instance=instance, filename=filename
        )

    def natural_key(self):
        return (self.label,)

    def new_document(
        self, file_object, label=None, description=None, language=None,
        _user=None
    ):
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )

        try:
            document = Document(
                description=description or '', document_type=self,
                label=label or file_object.name,
                language=language or setting_language.value
            )
            document._event_keep_attributes = ('_event_actor',)
            document._event_actor = _user
            document.save()
        except Exception as exception:
            logger.critical(
                'Unexpected exception while trying to create new document '
                '"%s" from document type "%s"; %s',
                label or file_object.name, self, exception
            )
            raise
        else:
            try:
                document_file = document.file_new(
                    file_object=file_object, filename=label, _user=_user
                )
            except Exception as exception:
                logger.critical(
                    'Unexpected exception while trying to create initial '
                    'file for document %s; %s',
                    label or file_object.name, exception
                )
                raise
            else:
                return document, document_file

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_document_type_created,
            'target': 'self'
        },
        edited={
            'event': event_document_type_edited,
            'target': 'self'
        }
    )

    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)


class DocumentTypeFilename(ExtraDataModelMixin, models.Model):
    """
    List of labels available to a specific document type for the
    quick rename functionality
    """
    document_type = models.ForeignKey(
        on_delete=models.CASCADE, related_name='filenames', to=DocumentType,
        verbose_name=_('Document type')
    )
    filename = models.CharField(
        db_index=True, max_length=128, verbose_name=_('Label')
    )
    enabled = models.BooleanField(default=True, verbose_name=_('Enabled'))

    class Meta:
        ordering = ('filename',)
        unique_together = ('document_type', 'filename')
        verbose_name = _('Quick label')
        verbose_name_plural = _('Quick labels')

    def __str__(self):
        return self.filename

    @method_event(
        event_manager_class=EventManagerMethodAfter,
        event=event_document_type_quick_label_deleted,
        target='document_type',
    )
    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_document_type_quick_label_created,
            'action_object': 'document_type',
            'target': 'self'
        },
        edited={
            'event': event_document_type_quick_label_edited,
            'action_object': 'document_type',
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)



import logging

from django.db import models
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from ..managers import DuplicatedDocumentManager

from .document_models import Document

__all__ = ('DuplicatedDocument',)
logger = logging.getLogger(name=__name__)


class DuplicatedDocument(models.Model):
    document = models.ForeignKey(
        on_delete=models.CASCADE, related_name='duplicates', to=Document,
        verbose_name=_('Document')
    )
    documents = models.ManyToManyField(
        to=Document, verbose_name=_('Duplicated documents')
    )
    datetime_added = models.DateTimeField(
        auto_now_add=True, db_index=True, verbose_name=_('Added')
    )

    objects = DuplicatedDocumentManager()

    class Meta:
        verbose_name = _('Duplicated document')
        verbose_name_plural = _('Duplicated documents')

    def __str__(self):
        return force_text(s=self.document)



from actstream.models import Action

from mayan.apps.documents.tests.base import GenericDocumentTestCase

from ..events import (
    event_ocr_document_version_content_deleted,
    event_ocr_document_version_submit, event_ocr_document_version_finish
)
from ..models import DocumentVersionPageOCRContent


class OCREventsTestCase(GenericDocumentTestCase):
    def test_document_content_deleted_event(self):
        Action.objects.all().delete()
        DocumentVersionPageOCRContent.objects.delete_content_for(
            document_version=self.test_document_version
        )

        # Get the oldest action
        action = Action.objects.order_by('-timestamp').last()

        self.assertEqual(
            action.target, self.test_document_version
        )
        self.assertEqual(
            action.action_object, self.test_document
        )
        self.assertEqual(
            action.verb, event_ocr_document_version_content_deleted.id
        )

    def test_document_version_submit_event(self):
        Action.objects.all().delete()
        self.test_document.submit_for_ocr()

        # Get the oldest action
        action = Action.objects.order_by('-timestamp').last()

        self.assertEqual(
            action.target, self.test_document.version_active
        )
        self.assertEqual(
            action.action_object, self.test_document
        )
        self.assertEqual(
            action.verb, event_ocr_document_version_submit.id
        )

    def test_document_version_finish_event(self):
        Action.objects.all().delete()
        self.test_document.submit_for_ocr()

        # Get the most recent action
        action = Action.objects.order_by('-timestamp').first()

        self.assertEqual(
            action.target, self.test_document.version_active
        )
        self.assertEqual(
            action.action_object, self.test_document
        )
        self.assertEqual(
            action.verb, event_ocr_document_version_finish.id
        )



from django.test import override_settings

from mayan.apps.documents.models.document_models import DocumentSearchResult
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.search import document_search
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.testing.tests.base import GenericViewTestCase
from mayan.apps.storage.utils import fs_cleanup, mkdtemp

from ..classes import SearchBackend, SearchModel
from ..permissions import permission_search_tools
from ..settings import setting_backend_arguments

from .mixins import SearchToolsViewTestMixin, SearchViewTestMixin


class AdvancedSearchViewTestCase(
    DocumentTestMixin, SearchViewTestMixin, GenericViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self.test_document_count = 4

        # Upload many instances of the same test document
        for i in range(self.test_document_count):
            self._upload_test_document()

        self.search_backend = SearchBackend.get_instance()

    def test_advanced_search_past_first_page(self):
        test_document_label = self.test_documents[0].label

        for document in self.test_documents:
            self.grant_access(
                obj=document, permission=permission_document_view
            )

        # Make sure all documents are returned by the search
        queryset = self.search_backend.search(
            search_model=document_search,
            query_string={'label': test_document_label},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), self.test_document_count)

        with self.settings(VIEWS_PAGINATE_BY=2):
            # Functional test for the first page of advanced results
            response = self._request_search_results_view(
                data={'label': test_document_label}, kwargs={
                    'search_model_name': document_search.get_full_name()
                }
            )

            # Total (1 - 2 out of 4) (Page 1 of 2)
            # 4 results total, 2 pages, current page is 1,
            # object in this page: 2
            self.assertEqual(response.status_code, 200)
            self.assertEqual(
                response.context['paginator'].object_list.count(), 4
            )
            self.assertEqual(response.context['paginator'].num_pages, 2)
            self.assertEqual(response.context['page_obj'].number, 1)
            self.assertEqual(
                response.context['page_obj'].object_list.count(), 2
            )

            # Functional test for the second page of advanced results
            response = self._request_search_results_view(
                data={'label': test_document_label, 'page': 2}, kwargs={
                    'search_model_name': document_search.get_full_name()
                }
            )
            # Total (3 - 4 out of 4) (Page 2 of 2)
            # 4 results total, 2 pages, current page is 1,
            # object in this page: 2
            self.assertEqual(response.status_code, 200)
            self.assertEqual(
                response.context['paginator'].object_list.count(), 4
            )
            self.assertEqual(response.context['paginator'].num_pages, 2)
            self.assertEqual(response.context['page_obj'].number, 2)
            self.assertEqual(
                response.context['page_obj'].object_list.count(), 2
            )


class SearchViewTestCase(
    DocumentTestMixin, SearchViewTestMixin, GenericViewTestCase
):
    def test_result_view_with_search_mode_in_data(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        response = self._request_search_results_view(
            data={
                'label': self.test_document.label,
                '_search_model_name': document_search.get_full_name()
            }
        )
        self.assertContains(
            response=response, status_code=200, text=self.test_document.label
        )


@override_settings(SEARCH_BACKEND='mayan.apps.dynamic_search.backends.whoosh.WhooshSearchBackend')
class SearchToolsViewTestCase(
    DocumentTestMixin, SearchToolsViewTestMixin, GenericViewTestCase
):
    def setUp(self):
        self.old_value = setting_backend_arguments.value
        super().setUp()
        self.document_search_model = SearchModel.get_for_model(
            instance=DocumentSearchResult
        )
        setting_backend_arguments.set(
            value={'index_path': mkdtemp()}
        )
        self.search_backend = SearchBackend.get_instance()

    def tearDown(self):
        fs_cleanup(
            filename=setting_backend_arguments.value['index_path']
        )
        setting_backend_arguments.set(value=self.old_value)
        super().tearDown()

    def test_search_backend_reindex_view_no_permission(self):
        self.search_backend.clear_search_model_index(
            search_model=self.document_search_model
        )
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        response = self._request_search_backend_reindex_view()
        self.assertEqual(response.status_code, 403)

        queryset = self.search_backend.search(
            search_model=self.document_search_model,
            query_string={'q': self.test_document.label},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 0)

    def test_search_backend_reindex_view_with_permission(self):
        self.search_backend.clear_search_model_index(
            search_model=self.document_search_model
        )
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )
        self.grant_permission(permission=permission_search_tools)

        response = self._request_search_backend_reindex_view()
        self.assertEqual(response.status_code, 302)

        queryset = self.search_backend.search(
            search_model=self.document_search_model,
            query_string={'q': self.test_document.label},
            user=self._test_case_user
        )
        self.assertNotEqual(queryset.count(), 0)



import logging

from django.apps import apps
from django.db import models
from django.db.models import Q, Value

from mayan.apps.acls.models import AccessControlList

from .classes import DuplicateBackend

logger = logging.getLogger(name=__name__)


class StoredDuplicateBackendManager(models.Manager):
    def scan_all(self):
        """
        Find duplicates by iterating over all documents and all backends.
        """
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )

        for document in Document.valid.all():
            self.scan_document(document=document, scan_children=False)

    def scan_document(self, document, scan_children=True):
        """
        Find duplicates by matching latest file checksums
        """
        if not document.file_latest:
            return None

        for backend_path, backend_class in DuplicateBackend.get_all():
            stored_backend, created = self.get_or_create(
                backend_path=backend_path
            )
            duplicates = stored_backend.get_backend_instance().process(document=document)

            if duplicates.exists():
                duplicates_entry, created = stored_backend.duplicate_entries.get_or_create(
                    document=document
                )
                duplicates_entry.documents.add(*duplicates)
            else:
                stored_backend.duplicate_entries.filter(document=document).delete()

            if scan_children:
                for document in duplicates:
                    self.scan_document(document=document, scan_children=False)


class DuplicateBackendEntryManager(models.Manager):
    def clean_empty_duplicate_lists(self):
        self.filter(documents=None).delete()

    def get_duplicated_documents(self, permission=None, user=None):
        DuplicateSourceDocument = apps.get_model(
            app_label='duplicates', model_name='DuplicateSourceDocument'
        )
        DuplicateTargetDocument = apps.get_model(
            app_label='duplicates', model_name='DuplicateTargetDocument'
        )

        target_document_queryset = DuplicateTargetDocument.valid.all()

        if permission:
            target_document_queryset = AccessControlList.objects.restrict_queryset(
                queryset=target_document_queryset, permission=permission,
                user=user
            )

        queryset = self.filter(documents__in=target_document_queryset).only(
            'document_id'
        ).values('document_id')

        return DuplicateSourceDocument.valid.filter(pk__in=queryset)

    def get_duplicates_of(self, document, permission=None, user=None):
        DuplicateTargetDocument = apps.get_model(
            app_label='duplicates', model_name='DuplicateTargetDocument'
        )

        queryset = DuplicateTargetDocument.valid.none()

        when_list = []

        for entry in self.filter(document=document).all():
            queryset = queryset | DuplicateTargetDocument.valid.filter(
                pk__in=entry.documents.only('pk').values('pk')
            )

            when_list.append(
                models.When(
                    Q(pk__in=entry.documents.only('pk').values('pk')),
                    then=Value(value=str(entry.stored_backend))
                )
            )

        queryset = queryset.annotate(
            backend=models.Case(
                *when_list, output_field=models.CharField()
            )
        )

        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset, user=user
            )

        return queryset



from django.utils.translation import ugettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_medium, worker_slow

queue_default = CeleryQueue(
    default_queue=True, label=_('Default'), name='default', worker=worker_medium
)
queue_tools = CeleryQueue(label=_('Tools'), name='tools', worker=worker_slow)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('document_signatures', '0010_auto_20191201_0146')
    ]

    operations = [
        migrations.RenameField(
            model_name='signaturebasemodel',
            old_name='document_version',
            new_name='document_file',
        ),
    ]
